package java.util.stream;

import java.util.Set;
import java.util.Objects;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.IntFunction;
import java.util.AbstractCollection;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Collection;
import java.util.Spliterator;

final class DistinctOps
{
    private DistinctOps() {
    }
    
    static <T> ReferencePipeline<T, T> makeRef(final AbstractPipeline<?, T, ?> abstractPipeline) {
        return new ReferencePipeline.StatefulOp<T, T>(abstractPipeline, StreamShape.REFERENCE, StreamOpFlag.IS_DISTINCT | StreamOpFlag.NOT_SIZED) {
             <P_IN> Node<T> reduce(PipelineHelper<T> pipelineHelper, Spliterator<P_IN> spliterator) {
                return Nodes.node((Collection)ReduceOps.makeRef(LinkedHashSet::new, HashSet::add, AbstractCollection::addAll).evaluateParallel(pipelineHelper, spliterator));
            }
            
            @Override
             <P_IN> Node<T> opEvaluateParallel(PipelineHelper<T> pipelineHelper, Spliterator<P_IN> spliterator, IntFunction<T[]> intFunction) {
                if (StreamOpFlag.DISTINCT.isKnown(pipelineHelper.getStreamAndOpFlags())) {
                    return pipelineHelper.evaluate(spliterator, false, intFunction);
                }
                if (StreamOpFlag.ORDERED.isKnown(pipelineHelper.getStreamAndOpFlags())) {
                    return this.reduce(pipelineHelper, spliterator);
                }
                AtomicBoolean atomicBoolean = new AtomicBoolean(false);
                ConcurrentHashMap<Object, Object> concurrentHashMap = new ConcurrentHashMap();
                ForEachOps.makeRef(key/* java.lang.Object, */ -> {
                    if (key == null) {
                        var_4_31.set(true);
                    }
                    else {
                        var_5_3A.putIfAbsent(key, Boolean.TRUE);
                    }
                    return;
                }, false).evaluateParallel(pipelineHelper, spliterator);
                ConcurrentHashMap.KeySetView<T, Object> keySet = concurrentHashMap.keySet();
                if (atomicBoolean.get()) {
                    keySet = new HashSet(keySet);
                    keySet.add(null);
                }
                return Nodes.node(keySet);
            }
            
            @Override
             <P_IN> Spliterator<T> opEvaluateParallelLazy(PipelineHelper<T> pipelineHelper, Spliterator<P_IN> spliterator) {
                if (StreamOpFlag.DISTINCT.isKnown(pipelineHelper.getStreamAndOpFlags())) {
                    return pipelineHelper.wrapSpliterator(spliterator);
                }
                if (StreamOpFlag.ORDERED.isKnown(pipelineHelper.getStreamAndOpFlags())) {
                    return this.reduce(pipelineHelper, spliterator).spliterator();
                }
                return new StreamSpliterators.DistinctSpliterator(pipelineHelper.wrapSpliterator(spliterator));
            }
            
            @Override
            Sink<T> opWrapSink(int n, Sink<T> obj) {
                Objects.requireNonNull(obj);
                if (StreamOpFlag.DISTINCT.isKnown(n)) {
                    return obj;
                }
                if (StreamOpFlag.SORTED.isKnown(n)) {
                    return new Sink.ChainedReference<T, T>(obj) {
                        boolean seenNull;
                        T lastSeen;
                        
                        @Override
                        public void begin(long n) {
                            this.seenNull = false;
                            this.lastSeen = null;
                            this.downstream.begin(-1L);
                        }
                        
                        @Override
                        public void end() {
                            this.seenNull = false;
                            this.lastSeen = null;
                            this.downstream.end();
                        }
                        
                        @Override
                        public void accept(T lastSeen) {
                            if (lastSeen == null) {
                                if (!this.seenNull) {
                                    this.seenNull = true;
                                    this.downstream.accept(this.lastSeen = null);
                                }
                            }
                            else if (this.lastSeen == null || !lastSeen.equals(this.lastSeen)) {
                                this.downstream.accept(this.lastSeen = lastSeen);
                            }
                        }
                    };
                }
                return new Sink.ChainedReference<T, T>(obj) {
                    Set<T> seen;
                    
                    @Override
                    public void begin(long n) {
                        this.seen = new HashSet();
                        this.downstream.begin(-1L);
                    }
                    
                    @Override
                    public void end() {
                        this.seen = null;
                        this.downstream.end();
                    }
                    
                    @Override
                    public void accept(T t) {
                        if (!this.seen.contains(t)) {
                            this.seen.add(t);
                            this.downstream.accept(t);
                        }
                    }
                };
            }
        };
    }
}
